#!/usr/bin/env python

"""
Common calendar client utilities.

Copyright (C) 2014, 2015 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from datetime import datetime
from imiptools.data import get_address, get_uri, get_window_end, uri_dict, uri_items, uri_values
from imiptools.period import update_freebusy
from imiptools.profile import Preferences
from imiptools.dates import format_datetime, get_default_timezone, \
                            to_timezone

def update_attendees(obj, attendees, removed):

    """
    Update the attendees in 'obj' with the given 'attendees' and 'removed'
    attendee lists. A list is returned containing the attendees whose
    attendance should be cancelled.
    """

    to_cancel = []

    existing_attendees = uri_values(obj.get_values("ATTENDEE") or [])
    added = set(attendees).difference(existing_attendees)

    if added or removed:
        attendees = uri_items(obj.get_items("ATTENDEE") or [])
        sequence = obj.get_value("SEQUENCE")

        if removed:
            remaining = []

            for attendee, attendee_attr in attendees:
                if attendee in removed:

                    # Without a sequence number, assume that the event has not
                    # been published and that attendees can be silently removed.

                    if sequence is not None:
                        to_cancel.append((attendee, attendee_attr))
                else:
                    remaining.append((attendee, attendee_attr))

            attendees = remaining

        if added:
            for attendee in added:
                attendee = attendee.strip()
                if attendee:
                    attendees.append((get_uri(attendee), {"PARTSTAT" : "NEEDS-ACTION", "RSVP" : "TRUE"}))

        obj["ATTENDEE"] = attendees

    return to_cancel

class Client:

    "Common handler and manager methods."

    default_window_size = 100

    def __init__(self, user, messenger=None):
        self.user = user
        self.messenger = messenger
        self.preferences = None

    def get_preferences(self):
        if not self.preferences and self.user:
            self.preferences = Preferences(self.user)
        return self.preferences

    def get_tzid(self):
        prefs = self.get_preferences()
        return prefs and prefs.get("TZID") or get_default_timezone()

    def get_window_size(self):
        prefs = self.get_preferences()
        try:
            return prefs and int(prefs.get("window_size")) or self.default_window_size
        except (TypeError, ValueError):
            return self.default_window_size

    def get_window_end(self):
        return get_window_end(self.get_tzid(), self.get_window_size())

    def is_sharing(self):
        prefs = self.get_preferences()
        return prefs and prefs.get("freebusy_sharing") == "share" or False

    def is_bundling(self):
        prefs = self.get_preferences()
        return prefs and prefs.get("freebusy_bundling") == "always" or False

    def is_notifying(self):
        prefs = self.get_preferences()
        return prefs and prefs.get("freebusy_messages") == "notify" or False

    # Common operations on calendar data.

    def update_participation(self, obj, partstat=None):

        """
        Update the participation in 'obj' of the user with the given 'partstat'.
        """

        attendee_attr = uri_dict(obj.get_value_map("ATTENDEE")).get(self.user)
        if not attendee_attr:
            return None
        if partstat:
            attendee_attr["PARTSTAT"] = partstat
        if attendee_attr.has_key("RSVP"):
            del attendee_attr["RSVP"]
        self.update_sender(attendee_attr)
        return attendee_attr

    def update_sender(self, attr):

        "Update the SENT-BY attribute of the 'attr' sender metadata."

        if self.messenger and self.messenger.sender != get_address(self.user):
            attr["SENT-BY"] = get_uri(self.messenger.sender)

class ClientForObject(Client):

    "A client maintaining a specific object."

    def __init__(self, obj, user, messenger=None):
        Client.__init__(self, user, messenger)
        self.set_object(obj)

    def set_object(self, obj):
        self.obj = obj
        self.uid = obj and self.obj.get_uid()
        self.recurrenceid = obj and self.obj.get_recurrenceid()
        self.sequence = obj and self.obj.get_value("SEQUENCE")
        self.dtstamp = obj and self.obj.get_value("DTSTAMP")

    def _update_freebusy(self, freebusy, periods, recurrenceid, transp=None):

        """
        Update the 'freebusy' collection with the given 'periods', indicating an
        explicit 'recurrenceid' to affect either a recurrence or the parent
        event.
        """

        update_freebusy(freebusy, periods,
            transp or self.obj.get_value("TRANSP") or "OPAQUE",
            self.uid, recurrenceid,
            self.obj.get_value("SUMMARY"),
            self.obj.get_value("ORGANIZER"))

    def update_freebusy(self, freebusy, periods, transp=None):

        """
        Update the 'freebusy' collection for this event with the given
        'periods'.
        """

        self._update_freebusy(freebusy, periods, self.recurrenceid, transp)

    def update_freebusy_for_participant(self, freebusy, periods, attr, for_organiser=False):

        """
        Update the 'freebusy' collection using the given 'periods', subject to
        the 'attr' provided for the participant, indicating whether this is
        being generated 'for_organiser' or not.
        """

        # Organisers employ a special transparency if not attending.

        if self.is_participating(attr, for_organiser):
            self.update_freebusy(freebusy, periods,
                transp=self.get_overriding_transparency(attr, for_organiser))
        else:
            self.remove_from_freebusy(freebusy)

    def is_participating(self, attr, as_organiser=False):
        return as_organiser or not attr or attr.get("PARTSTAT") != "DECLINED"

    def get_overriding_transparency(self, attr, as_organiser=False):
        return as_organiser and not (attr and attr.get("PARTSTAT")) and "ORG" or None

    def update_dtstamp(self):

        "Update the DTSTAMP in the current object."

        dtstamp = self.obj.get_utc_datetime("DTSTAMP")
        utcnow = to_timezone(datetime.utcnow(), "UTC")
        self.obj["DTSTAMP"] = [(format_datetime(dtstamp > utcnow and dtstamp or utcnow), {})]

    def set_sequence(self, increment=False):

        "Update the SEQUENCE in the current object."

        sequence = self.obj.get_value("SEQUENCE") or "0"
        self.obj["SEQUENCE"] = [(str(int(sequence) + (increment and 1 or 0)), {})]

# vim: tabstop=4 expandtab shiftwidth=4
